package com.sillycycle.bagleyd.life4d;

/* Code for selection spliced in is due to Alex Shvets */
/* http://home.comcast.net/~shvets/Programs/Examples/GameLife/life2.html */
// Life4DMatrix.java

import java.io.Externalizable;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Comparator;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import com.sillycycle.bagleyd.util.Box4D;
import com.sillycycle.bagleyd.util.CellInfo;

public class Life4DMatrix implements Externalizable, Comparator<CellInfo> {
	private static final long serialVersionUID = 42L;

	private int columns, rows, stacks, zones;
	private int area, volume, hypervolume;

	private byte[] cells;
	public transient List<CellInfo> list, deadList;

	// must exist for deserialization process
	public Life4DMatrix() {
	}

	public Life4DMatrix(int columns, int rows, int stacks, int zones) {
		this.columns = columns;
		this.rows = rows;
		this.stacks = stacks;
		this.zones = zones;
		area = columns * rows;
		volume = area * stacks;
		hypervolume = volume * zones;
		list = new LinkedList<>();
		deadList = new LinkedList<>();
		init();
	}

	@Override
	public int compare(CellInfo a, CellInfo b) {
		if (a.getDistance() < b.getDistance())
			return 1;
		if (a.getDistance() > b.getDistance())
			return -1;
		return 0;
	}

	public void init() {
		cells = new byte[hypervolume];
		clear();
	}

	public void clear() {
		for (int position = 0; position < hypervolume; position++)
			cells[position] = (byte) 0; // Empty
	}

	public void clearLiveCells() {
		for (Iterator<CellInfo> iterator = list.iterator();
				iterator.hasNext();) {
			CellInfo cellInfo = iterator.next();
			int position = cellInfo.getPosition();
			cells[position] = 0;
			iterator.remove();
		}
	}

	// O(n)
	public void clearCell(int position) {
		for (Iterator<CellInfo> iterator = list.iterator();
				iterator.hasNext();) {
			CellInfo cellInfo = iterator.next();
			int p = cellInfo.getPosition();
			if (position == p)
				iterator.remove();
			//break; // commented to clear duplications
		}
	}

	public int getRows() {
		return rows;
	}

	public int getColumns() {
		return columns;
	}

	public int getStacks() {
		return stacks;
	}

	public int getZones() {
		return zones;
	}

	public byte getCell(int position) {
		byte value = cells[position];

		if (value == -128)
			return 0;
		else if (value < 0)
			return (byte) -value;
		else
			return value;
	}

	public boolean getCellBit(int position) {
		return (cells[position] < 0);
	}

	public void setArrayCell(int position, byte cell) {
		cells[position] = cell;
	}

	public void setCell(int position, byte cell) {
		if (cells[position] == 0) {
			list.add(new CellInfo(position, cell));
		} else if (cell == 0) {
			for (Iterator<CellInfo> iterator = list.iterator();
					iterator.hasNext();) {
				CellInfo cellInfo = iterator.next();
				int p = cellInfo.getPosition();
				if (position == p)
					iterator.remove();
				//break; // commented to clear duplications
			}
		 } else {
			for (CellInfo cellInfo : list) {
				int p = cellInfo.getPosition();
				if (position == p) {
					cellInfo.setCell(cell);
				}
			}
		}
		cells[position] = cell;
	}

	public void setDeadCell(int position) {
		if (cells[position] == 0) {
			setCellBit(position);
			CellInfo cellInfo = new CellInfo(position);
			deadList.add(cellInfo);
//System.out.println(" " + position + " " + getCellBit(position));
		}
	}

	public void setCellBit(int position) {
		byte value = cells[position];

		if (value == 0)
			value = -128;
		else if (value > 0)
			value = (byte) -value;
		cells[position] = value;
	}

	public void resetCellBit(int position) {
		byte value = cells[position];

		if (value == -128)
			value = 0;
		else if (value < 0)
			value = (byte) -value;
		cells[position] = value;
	}

	public byte getCell(int column, int row, int stack, int zone) {

		if ((zone >= 0) && (zone < zones) &&
				(stack >= 0) && (stack < stacks) &&
				(row >= 0) && (row < rows) &&
				(column >= 0) && (column < columns))
			return getCell(zone * volume + stack * area + row * columns + column);
		return 0;
	}

	// To be used so do not add to dead list for cells already added
	public boolean getCellBit(int column, int row, int stack, int zone) {
		if ((zone >= 0) && (zone < zones) &&
				(stack >= 0) && (stack < stacks) &&
				(row >= 0) && (row < rows) &&
				(column >= 0) && (column < columns))
			return getCellBit(zone * volume + stack * area + row * columns + column);
		return true;
	}

	public void setCell(int column, int row, int stack, int zone, byte cell) {
		if ((zone >= 0) && (zone < zones) &&
				(stack >= 0) && (stack < stacks) &&
				(row >= 0) && (row < rows) &&
				(column >= 0) && ( column < columns)) {
			int position = stack * area + row * columns + column;
			if (cell == 0) {
				clearCell(position);
			} else {
				setCell(position, cell);
			}
		}
	}

	// multiple loops are ok for SubMatrix stuff, time not an issue

	public Life4DMatrix getSubMatrix(Box4D b) {
		Life4DMatrix selection = new Life4DMatrix(b.width, b.height, b.depth, b.zone);

		for (int l = 0; l < b.zone; l++)
		for (int k = 0; k < b.depth; k++)
		for (int j = 0; j < b.height; j++)
		for (int i = 0; i < b.width; i++) {
			selection.setCell(i, j, k, l,
				getCell(i + b.x, j + b.y, k + b.z, l + b.w));
		}
		return selection;
	}

	public void setSubMatrix(Life4DMatrix selection, Box4D b) {
		for (int l = 0; l < b.zone; l++)
		for (int k = 0; k < b.depth; k++)
		for (int j = 0; j < b.height; j++)
		for (int i = 0; i < b.width; i++) {
			try {
				setCell(i + b.x, j + b.y, k + b.z, l + b.w,
					selection.getCell(i, j, k, l));
			} catch (ArrayIndexOutOfBoundsException e) {
				// ignore
			}
		}
	}

	public void clearSubMatrix(Box4D b) {
		for (int l = b.w; l < b.w + b.zone; l++)
		for (int k = b.z; k < b.z + b.depth; k++)
		for (int j = b.y; j < b.y + b.height; j++)
		for (int i = b.x; i < b.x + b.width; i++) {
			cells[l * volume + k * area + j * columns + i] = 0; //Empty
		}
	}

	/* The center point of rectangle (more or less) is used
	 * to rotate about.
	 */
	void rotate(Life4DMatrix selection, Box4D b,
			int plane, boolean cw) {
		for (int l = 0; l < b.zone; l++)
		for (int k = 0; k < b.depth; k++)
		for (int j = 0; j < b.height; j++)
		for (int i = 0; i < b.width; i++) {
			if (plane == 0) {
				if (cw)
					setCell(b.height - j - 1 + b.x,
						i + b.y,
						k + b.z,
						l + b.w,
						selection.getCell(i, j, k, l));
				else
					setCell(j + b.x,
						b.width - i - 1 + b.y,
						k + b.z,
						l + b.w,
						selection.getCell(i, j, k, l));
			}
		}
	}

	/* Wall means you reflecting in a wall mirror,
	 * else it means you are reflecting off ground.
	 */
	void reflect(Life4DMatrix selection, Box4D b,
			int plane, boolean wall) {
		for (int l = 0; l < b.zone; l++)
		for (int k = 0; k < b.depth; k++)
		for (int j = 0; j < b.height; j++)
		for (int i = 0; i < b.width; i++) {
			if (plane == 0) {
				if (wall)
					setCell(b.width - i - 1 + b.x,
						j + b.y,
						k + b.z,
						l + b.w,
						selection.getCell(i, j, k, l));
				else
					setCell(i + b.x,
						b.height - j - 1 + b.y,
						k + b.z,
						l + b.w,
						selection.getCell(i, j, k, l));
			}
		}
	}


	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder();

		for (int l = 0; l < zones; l++) {
		for (int k = 0; k < stacks; k++) {
		for (int j = 0; j < rows; j++) {
			for (int i = 0; i < columns; i++)
				sb.append(cells[l * volume + k * area + j * columns + i]);
			sb.append("\n");
		}
		sb.append("\n");
		}
		sb.append("\n");
		}
		sb.append("\n");
		return sb.toString();
	}


	// implements Externalizable

	@Override
	public void writeExternal(ObjectOutput out) throws IOException {
		out.writeByte(columns);
		out.writeByte(rows);
		out.writeByte(stacks);
		out.writeByte(zones);
		for (int l = 0; l < zones; l++)
		for (int k = 0; k < stacks; k++)
		for (int j = 0; j < rows; j++)
		for (int i = 0; i < columns; i++) {
			if (cells[l * volume + k * area + j * columns + i] == 1) {
				out.writeByte(i);
				out.writeByte(j);
				out.writeByte(k);
				out.writeByte(l);
			}
		}
	}

	@Override
	public void readExternal(ObjectInput in)
			throws IOException, ClassNotFoundException {
		columns = in.readByte();
		rows = in.readByte();
		stacks = in.readByte();
		zones = in.readByte();
		init();
		while (true) {
			try {
				int i = in.readByte();
				int j = in.readByte();
				int k = in.readByte();
				int l = in.readByte();
				cells[l * volume + k * area + j * columns + i] = 1;
			} catch (IOException e)	{
				break;
			}
		}
	}

	public static void main(String[] args) {
		System.out.println("Constructing object:");
		Life4DMatrix m = new Life4DMatrix(9, 8, 7, 6);
		m.setCell(2, 3, 2, 1, (byte) 1);
		m.setCell(3, 3, 2, 1, (byte) 1);
		m.setCell(4, 3, 2, 1, (byte) 1);

		m.setCell(3, 5, 2, 1, (byte) 1);
		m.setCell(2, 3, 3, 1, (byte) 1);
		m.setCell(3, 3, 3, 1, (byte) 1);
		m.setCell(4, 3, 3, 1, (byte) 1);
		m.setCell(4, 4, 3, 1, (byte) 1);
		m.setCell(3, 5, 3, 1, (byte) 1);
		Life4DMatrix n;

		System.out.println("Saving object:");
		try (ObjectOutputStream out = new ObjectOutputStream(
				new FileOutputStream("junk.out"))) {
			out.writeObject(m);
		} catch (Exception e) {
			e.printStackTrace();
		}
		// Now get it back
		System.out.println("Recovering object:");
		try (ObjectInputStream in = new ObjectInputStream(
				new FileInputStream("junk.out"))) {
			n = (Life4DMatrix)in.readObject();
			System.out.println(n.toString());
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
}
